package org.dbxml.core.objects;

/*
 * dbXML License, Version 1.0
 *
 *
 * Copyright (c) 1999-2001 The dbXML Group, L.L.C.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        The dbXML Group (http://www.dbxml.com/)."
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 *
 * 4. The names "dbXML" and "The dbXML Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    info@dbxml.com.
 *
 * 5. Products derived from this software may not be called "dbXML",
 *    nor may "dbXML" appear in their name, without prior written
 *    permission of The dbXML Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE DBXML GROUP OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: Reflectable.java,v 1.7 2001/05/23 00:35:11 tom Exp $
 */

import org.w3c.dom.*;

/**
 * Reflectable is the base interface for XMLObjects, allowing them to expose a
 * Reflector for performing method introspection and execution against the
 * Object.
 * <br><br>
 * dbXML will expose all public methods that match a valid signature.
 * The rules of this signature follow:
 * <pre>
 *    Result/Params: void                              (result only)
 *                   org.w3c.dom.Element               (result only)
 *                   org.w3c.dom.Document              (result only)
 *                   org.w3c.dom.DocumentFragment      (result only)
 *                   native types (long, int, etc...)
 *                   java.lang.String
 *                   org.dbxml.core.objects.Variant
 *
 *    *NOTE: String, Variant, and all of the Java native types (long,
 *           int, etc...) can also be passed as arrays of those types.
 * </pre>
 * XMLObjects need to expose meta data so that a URI Query String (or any
 * other type of parameter set can be mapped to a method's arguments.  This
 * is accomplished through the getReflector() method.
 * <br><br>
 * A typical URI for an Reflectable might be:
 * <pre>
 *    /Database/Collection/1234567/XMLObject/Method(Blah, Blah)
 *    /Database/Collection/1234567/XMLObject/Method?red=Blah&green=Blah
 * </pre>
 * Both forms will create a new Reflectable if neccessary, set any
 * required context information, and call a method in that object.  The
 * first form passes the two arguments as-is, and the second maps the 'red'
 * parameter to the red argument, and the blue parameter to the blue
 * argument (regardless of method ordering).
 */

public interface Reflectable {
   static final String[] NO_PARAMS = new String[0];

   /**
    * getReflector returns the Reflectable Object's Reflector instance.
    *
    * @return The Object's Reflector
    */
   Reflector getReflector();

   /**
    * queryInterface allows a client to query the Reflectable Object for
    * exposed methods.
    *
    * @return The Object's interface
    */
   Document queryInterface();
}
